/*======================================================================

    Dummy device for "remote" kernel debugging with gdb

    Copyright (C) 1995 by David Hinds, dhinds@allegro.stanford.edu
    Single-step/breakpoint modifications
    Copyright (C) 1995  by John Heidemann <johnh@isi.edu>.
    Upgraded to 2.1.55.  Minor bug fixes.  Automatic removal of
    breakpoints that are invoked with interrupts disabled.
    1997/09/15 Keith Owens <kaos@ocs.com.au>

    $Id: xkdebug.c,v 1.2 1995/10/16 20:53:10 johnh Exp $
    Derived from:
    kdebug.c 1.6 1995/06/07 06:05:50 (David Hinds)

======================================================================*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/malloc.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/vmalloc.h>

#include <asm/hardirq.h>
#include <asm/ptrace.h>
#include <asm/segment.h>
#include <asm/uaccess.h>

/*====================================================================*/

static long debug_read(struct inode *inode, struct file *file,
		      char *buf, unsigned long count);
static long debug_write(struct inode *inode, struct file *file,
		       const char *buf, unsigned long count);
static int debug_ioctl(struct inode *inode, struct file *file,
		       u_int cmd, u_long arg);
static int debug_open(struct inode *inode, struct file *file);
static int debug_release(struct inode *inode, struct file *file);

static struct file_operations debug_fops = {
    NULL,		/* lseek */
    debug_read,
    debug_write,
    NULL,		/* readdir */
    NULL,		/* poll */
    debug_ioctl,
    NULL,		/* mmap */
    debug_open,
    debug_release,
    NULL		/* fsync */


};
    
/* Major device # for debug pseudo-device */
static int major_dev = 0;

/* Device in use? */
static int is_open = 0;

static int log_packets = 0;   /* report packets as they go by? */

/* Input and output buffers */
#define BUFSIZE 512
static u_char *inq, *outp, *outq;
static u_char inbuf[BUFSIZE], outbuf[BUFSIZE];

/*====================================================================*/

#define STACKSIZE 4096
/* Registers must match gdb's preferences. */
enum registers {
    R_EAX, R_ECX, R_EDX, R_EBX, R_XESP, R_EBP, R_XESI, R_EDI,
    R_EIP, R_EFLAGS, R_XCS, R_XSS, R_XDS, R_XES, R_XFS, R_XGS,
};
#define NREG ((int)R_XGS + 1)

static int float_stack = 0;
static u_char stack[STACKSIZE];

#define EFLAGS_TRACE_FLAG 0x100
/* Map to convert gdb <-> kernel regs */
#define KREGS_OFFSET(f) ((char*)(&((struct pt_regs*)NULL)->f)-(char*)NULL)
#define COMPUTE_KREGS_PLACE(kregs,off) (&(((u_char*) (kregs))[off]))
static struct {
    int offset;
    int size;
} kregs_info[] = {
    { KREGS_OFFSET(eax), 4 },
    { KREGS_OFFSET(ecx), 4 },
    { KREGS_OFFSET(edx), 4 },
    { KREGS_OFFSET(ebx), 4 },
    { KREGS_OFFSET(esp), 4 },
    { KREGS_OFFSET(ebp), 4 },
    { KREGS_OFFSET(esi), 4 },
    { KREGS_OFFSET(edi), 4 },
    { KREGS_OFFSET(eip), 4 },
    { KREGS_OFFSET(eflags), 4 },
    { KREGS_OFFSET(xcs), 4 },
    { KREGS_OFFSET(xss), 4 },
    { KREGS_OFFSET(xds), 4 },
    { KREGS_OFFSET(xes), 4 },
    { 0, 4 /* FS is a special case */ },
    { 0, 4 /* GS is a special case */ },
};

static struct wait_queue *kdebug_debugees = NULL;
static struct wait_queue *kdebug_debugger = NULL;
static struct pt_regs *current_debugee_regs = NULL;

/*====================================================================*/

/* If a breakpoint is executed while the cpu is running with interrupts
 * disabled, it is not possible to wake up the debugger.  This used to
 * cause an immediate kernel hang.  So we maintain our own list of
 * breakpoints and the original opcode.  If a breakpoint triggers while
 * interrupts are disabled, removed the original opcode and redrive
 * the instruction.  The kernel does not quite match what gdb thinks is
 * going on but it is better than a dead machine.
 */

typedef struct original {
    struct list_head head;
    caddr_t eip;
    u_char state;	/* 0 free, 1 breakpoint on,
			   2 disabled by interrupt check */
    u_char opcode;
} ORIGINAL;

ORIGINAL o_base;

#define INT3 0xcc
static const char INT3str[] = "cc";

static ORIGINAL * find_breakpoint(caddr_t eip)
{
    ORIGINAL *p = (ORIGINAL *) o_base.head.next;
    while (p != &o_base) {
	if (p->eip == eip)
	    return p;
	p = (ORIGINAL *) (p->head.next);
    }
    return NULL;
}

static int record_breakpoint(caddr_t eip)
{
    ORIGINAL *p = find_breakpoint(eip);
    if (!p) {
	p = kmalloc(sizeof(ORIGINAL), GFP_KERNEL);
	if (!p) {
	    printk(KERN_ERR "xkdebug cannot kmalloc new ORIGINAL\n");
	    return 1;
	}
	list_add(&p->head, &o_base.head);
    }
    p->eip = eip;
    p->state = 1;
    p->opcode = *((u_char *) eip);
    return 0;
}

static int remove_breakpoint(caddr_t eip)
{
    ORIGINAL *p = find_breakpoint(eip);
    if (!p)
	return 1;
    if (p->state != 1)
	return 1;
    *((u_char *) eip) = p->opcode;
    p->state = 2;
    printk(KERN_WARNING
    	"xkdebug breakpoint removed at %p, interrupts disabled\n",
    	eip);
    return 0;
}

/*====================================================================*/

static u_char *stack_base = stack;

static void update_stack(u_long esp)
{
    u_char *addr = (u_char *)esp;
    if ((addr < stack_base) || (addr > stack_base+STACKSIZE))
	stack_base = addr-STACKSIZE/2;
}

static u_char *map_stack(u_char *addr)
{
    if ((addr > stack_base) && (addr < stack_base+STACKSIZE))
	return addr + (stack - stack_base);
    else
	return addr;
}

/*====================================================================*/

static u_long call;

static void call_kernel(void)
{
#if 0
    /* Extract the target address from the call dummy */
    call = *(u_long *)(reg[R_EIP]+1) + reg[R_EIP] + 5;

    asm("movl %eax, _save\n movl _reg, %eax");
    asm("movl %ecx, _save+4\n movl _reg+4, %ecx");
    asm("movl %edx, _save+8\n movl _reg+8, %edx");
    asm("movl %ebx, _save+12\n movl _reg+12, %ebx");
    asm("movl %esp, _save+16\n movl _reg+16, %esp");
    asm("movl %ebp, _save+20\n movl _reg+20, %ebp");
    asm("movl %esi, _save+24\n movl _reg+24, %esi");
    asm("movl %edi, _save+28\n movl _reg+28, %edi");
    
    asm("movl _call, %eax\n call *%eax");
    
    asm("movl %eax, _reg\n movl _save, %eax");
    asm("movl %ecx, _reg+4\n movl _save+4, %ecx");
    asm("movl %edx, _reg+8\n movl _save+8, %edx");
    asm("movl %ebx, _reg+12\n movl _save+12, %ebx");
    asm("movl %esp, _reg+16\n movl _save+16, %esp");
    asm("movl %ebp, _reg+20\n movl _save+20, %ebp");
    asm("movl %esi, _reg+24\n movl _save+24, %esi");
    asm("movl %edi, _reg+28\n movl _save+28, %edi");
    
    /* Pretend we stopped at the call dummy breakpoint */
    reg[R_EIP] += 6;
#else /* ! 0 */
    /*
     * I take responsibility for breaking call_kernel.
     * It will come back.
     * --johnh 16-Oct-95
     */
    printk(KERN_ERR "call_kernel is currently broken.\n");
#endif /* 0 */
}

/*====================================================================*/

static u_char hexchars[] = "0123456789abcdef";

static int hex(int ch)
{
    if ((ch >= '0') && (ch <= '9')) return (ch-'0');
    if ((ch >= 'a') && (ch <= 'f')) return (ch-'a'+10);
    if ((ch >= 'A') && (ch <= 'F')) return (ch-'A'+10);
    /* Since noone checks errors, print. */
    printk(KERN_ERR "hex: bogus hex character %c\n", ch);
    return -1;
}

/*====================================================================*/

static void put_debug(char ch)
{
    *outq = ch;
    outq++;
    if (outq == outbuf + BUFSIZE)
	outq = outbuf;
    if (outp == outq)
	printk(KERN_ERR "kdebug: output buffer overflow!\n");
}

/*====================================================================*/

static void put_packet(char *buf)
{
    u_char checksum;

    if (log_packets)
	printk(KERN_DEBUG "put_packet('%s')\n", buf);
    put_debug('$');
    checksum = 0;
    while (*buf != '\0') {
	put_debug(*buf);
	checksum += *buf;
	buf++;
    }
    put_debug('#');
    put_debug(hexchars[checksum >> 4]);
    put_debug(hexchars[checksum % 16]);
}

/*====================================================================*/

static int verify_addr(u_char *addr, u_long len)
{
#ifdef HAS_VVERIFY
    if ((u_long)addr < PAGE_SIZE)
	return -1;
    else if ((void *)addr > high_memory)
	return vverify(addr, len);
    else
	return 0;
#else
    if ((u_long)addr < PAGE_SIZE)
	return -1;
    else
	return 0;
#endif
}

void read_registers(char *dest)
{
    u_char *src, byte;
    int i, j, dummy_int;
    void *dummy_src;

    for (i = R_EAX; i <= R_XGS; i++) {
	switch (i) {
	case R_XESP:
	    dummy_src = (void*)&current_debugee_regs->esp;
	    src = (u_char*)&dummy_src;
	    break;
	case R_XSS:
	    dummy_src = (u_char*)KERNEL_DS;
	    src = (u_char*)&dummy_src;
	    break;
	/* FS/GS asm code stolen from _fs() in traps.c */
	case R_XFS:
	    dummy_int = ({ register unsigned short __res;
		    __asm__("mov %%fs,%%ax":"=a" (__res):);
		    __res;});
	    src = (u_char*)&dummy_int;
	    break;
	case R_XGS:
	    dummy_int = ({ register unsigned short __res;
		    __asm__("mov %%gs,%%ax":"=a" (__res):);
		    __res;});
	    src = (u_char*)&dummy_int;
	    break;
	default:
	    src = COMPUTE_KREGS_PLACE(current_debugee_regs, kregs_info[i].offset);
	    break;
	};
	for (j = 0; j < 4; j++) {
	    byte = (j < kregs_info[i].size) ? *src++ : 0;
	    *dest++ = hexchars[(byte >> 4) & 0xf];
	    *dest++ = hexchars[byte & 0xf];
	};
    };
    *dest++ = '\0';
}

char *write_register(int reg_num, char *src)
{
    int j;
    u_char *dest  = COMPUTE_KREGS_PLACE(current_debugee_regs,
					kregs_info[reg_num].offset);
    void *dummy_dest;

    if (reg_num == R_XESP || reg_num == R_XSS ||
        reg_num == R_XFS || reg_num == R_XGS) {
	dest = (u_char*)&dummy_dest;
	/* ignore copy */
	/* xxx: should report changes as an error */
    };
    for (j = 0; j < 4; j++) {
	if (j < kregs_info[reg_num].size)
	    *dest++ = (hex(src[0]) << 4) + hex(src[1]);
	src += 2;
    };
    if (reg_num == R_EIP && *dest == 0xe5e5e5e5) {
	float_stack = 1;
	if (float_stack)
	    update_stack(*(u_long*)dest);
    };
    return src;
}

void write_registers(char *src)
{
    int i;

    for (i = R_EAX; i <= R_XGS; i++) {
	src = write_register(i, src);
    };
}

/*====================================================================*/

static void handle_packet(void)
{
    char *ptr;
    u_char *inp, *start, *addr;
    u_char buf[BUFSIZE];
    u_long i, len, reg_num;
    
    inp = inbuf;
    while (*inp != '$') inp++;

    start = ++inp;
    for (; *inp != '#'; inp++) ;
    *inp = '\0';

    if (log_packets)
	printk(KERN_DEBUG "handle_packet('%s')\n", start);
    
    put_debug('+');
    if (start[2] == ':') {
	put_debug(start[0]);
	put_debug(start[1]);
	start += 3;
    }

    buf[0] = '\0';
    switch (start[0]) {
    case '?': /* last signal */
	strcpy(buf, "S05");
	break;
    case 'd': /* toggle debug flag */
	break;
    case 'g': /* read registers */
	read_registers(buf);
	break;
    case 'G': /* write registers */
	write_registers(start+1);
	strcpy(buf, "OK");
	break;
    case 'P': /* write individual register */
	ptr = start + 1;
	reg_num = simple_strtoul(ptr, &ptr, 16);
	if (*ptr++ != '=')
	    break;   /* mis-formed packet */
	write_register(reg_num, ptr);
	strcpy(buf, "OK");
        break;
    case 'm': /* read memory */
	strcpy(buf, "E01");
	ptr = start + 1;
	addr = (caddr_t)simple_strtoul(ptr, &ptr, 16);
	if (float_stack)
	    addr = map_stack(addr);
	if (*ptr == ',') {
	    ptr++;
	    len = simple_strtoul(ptr, &ptr, 16);
	    if (verify_addr(addr, len))
		break;
	    for (i = 0; i < len; i++) {
		buf[i<<1] = hexchars[addr[i] >> 4];
		buf[(i<<1)+1] = hexchars[addr[i] % 16];
	    }
	    buf[len*2] = '\0';
	}
	break;
    case 'M': /* write memory */
	strcpy(buf, "E02");
	ptr = start + 1;
	addr = (caddr_t)simple_strtoul(ptr, &ptr, 16);
	if (float_stack)
	    addr = map_stack(addr);
	if (*ptr == ',') {
	    ptr++;
	    len = simple_strtoul(ptr, &ptr, 16);
	    if (verify_addr(addr, len))
		break;
	    if (*ptr == ':') {
		ptr++;
		if (len == 1) {
		    if (strncmp(ptr, INT3str, 2) == 0) {
			if (record_breakpoint(addr))
			    break;
		    }
		    else {
			ORIGINAL *p = find_breakpoint(addr);
			u_char c = (hex(ptr[0])<<4) + hex(ptr[1]);
			if (p && p->state == 1 && p->opcode == c)
			    p->state = 0;	/* breakpoint disabled */
		    }
		}
		for (i = 0; i < len; i++, ptr += 2)
		    addr[i] = (hex(ptr[0])<<4) + hex(ptr[1]);
		strcpy(buf, "OK");
	    }
	}
	break;
    case 'c': /* continue */
	if (stack_base != stack) {
	    strcpy(buf, "S0b");
	    break;
	}
	wake_up(&kdebug_debugees);
	interruptible_sleep_on(&kdebug_debugger);
	strcpy(buf, "S05");
	break;
    case 's': /* step */
	/* Set the step bit, step, then clear it. */
	current_debugee_regs->eflags |= EFLAGS_TRACE_FLAG;
	wake_up(&kdebug_debugees);
	interruptible_sleep_on(&kdebug_debugger);
	current_debugee_regs->eflags &= ~EFLAGS_TRACE_FLAG;
	strcpy(buf, "S05");
	break;
    case 'k': /* kill */
	break;
    }
    put_packet(buf);
    
    inq = inbuf;
}

/*====================================================================*/

static long debug_read(struct inode *inode, struct file *file,
		      char *buf, unsigned long count)
{
    int i;

    for (i = 0; i < count; i++) {
	if (outp == outq) break;
	put_user(*outp, buf);
	buf++; outp++;
	if (outp == outbuf+BUFSIZE)
	    outp = outbuf;
    }
    return i;
}

/*====================================================================*/

static enum { HEAD, BODY, CSUM1, CSUM2 } state = HEAD;

static long debug_write(struct inode *inode, struct file *file,
		       const char *buf, unsigned long count)
{
    int i;
    char ch;
    
    for (i = 0; i < count; i++) {
	if (inq == inbuf+BUFSIZE) break;
	get_user(ch, (unsigned char *) buf);
	*inq = ch;
	inq++; buf++;
	switch (state) {
	case HEAD:
	    if (ch == '$') state = BODY;
	    break;
	case BODY:
	    if (ch == '#') state = CSUM1;
	    break;
	case CSUM1:
	    state = CSUM2;
	    break;
	case CSUM2:
	    state = HEAD;
	    handle_packet();
	    break;
	}
    }
    return i;
}

/*====================================================================*/

static int debug_ioctl(struct inode *inode, struct file *file,
		       u_int cmd, u_long arg)
{
    return 0;
}

/*====================================================================*/

static void generate_fake_regs(void)
{
    static struct pt_regs fake_regs;

#if 0
    /* xxx: assume that these values are always constant. */
    reg[R_XESP] = save[R_XESP] = (u_long)(stack+STACKSIZE/2);
    asm("movw %xcs, _reg+40");
    asm("movw %xss, _reg+44");
    asm("movw %xds, _reg+48");
    asm("movw %xes, _reg+52");
    asm("movw %xfs, _reg+56");
    asm("movw %xgs, _reg+60");
#endif /* 0 */
    
    /*
     * To make gdb happy before the first "continue"
     * we need to provide registers.
     */
    current_debugee_regs = &fake_regs;
}


static int debug_open(struct inode *inode, struct file *file)
{
    if (!suser())
	return -EPERM;
    if (is_open)
	return -EBUSY;
    is_open = 1;
    inq = inbuf;
    outp = outq = outbuf;
    INIT_LIST_HEAD(&o_base.head);

    generate_fake_regs();
    put_packet("S02");   /* sigint---just for variety */
    MOD_INC_USE_COUNT;
    return 0;
}

/*====================================================================*/

static int debug_release(struct inode *inode, struct file *file)
{
    struct list_head *old;
    /*
     * Wake up anyone sleeping.
     * Presumably gdb has removed the breakpoints.
     */
    wake_up(&kdebug_debugees);
    while (!list_empty(&o_base.head)) {
    	old = o_base.head.next;
	list_del(old);
	kfree(old);
    }
    is_open = 0;
    MOD_DEC_USE_COUNT;
    return 0;
}

/*====================================================================*/

static asmlinkage int (*old_int1_hook)(struct pt_regs * regs, long error_code, int intr);
static asmlinkage int (*old_int3_hook)(struct pt_regs * regs, long error_code, int intr);

static asmlinkage int kdebug_do_ints(struct pt_regs * regs, long error_code, int intr)
{
    /* Yes. Suspend the trapee and wake up the trapper. */
    /* xxx: race condition on the global variable current_debugee_regs. */

    /* Avoid hung machines when breakpoint triggers with interrupts disabled */
    if (local_irq_count[smp_processor_id()]) {
	caddr_t eip = (caddr_t) regs->eip - 1;	/* back one byte */
	if (*((u_char *) eip) == INT3) {
	    if (remove_breakpoint(eip) == 0) {
		regs->eip = (long) eip;
		return 1;
	    }
	}
	panic("xkdebug breakpoint in interrupt, cannot recover");
    }

    current_debugee_regs = regs;
    if (kdebug_debugger)
	wake_up(&kdebug_debugger);
    sleep_on(&kdebug_debugees);
    return 1;   /* handled */
}

int init_module(void)
{
    major_dev = register_chrdev(0, "xkdebug", &debug_fops);
    if (major_dev == 0) {
	printk(KERN_ERR "unable to grab device # for debug driver\n");
	return -1;
    }
    old_int1_hook = kernel_int1_hook;
    kernel_int1_hook = kdebug_do_ints;
    old_int3_hook = kernel_int3_hook;
    kernel_int3_hook = kdebug_do_ints;
    return 0;
}

void cleanup_module(void)
{
    kernel_int1_hook = old_int1_hook;
    kernel_int3_hook = old_int3_hook;
    if (major_dev != 0)
	unregister_chrdev(major_dev, "xkdebug");
}

/*
 * Local Variables:
 * c-basic-offset:4
 * End:
 */
